@@ -115,11 +115,3 @@ class WidgetHandler(CommHandler):
 
     async def handle_comm_msg(self, msg: messages.CommMsg):
         self.state.update(msg.content.data["state"])
-
-    # Below methods are only used with Output widgets
-    # What will happen is that a special Handler for execute requests will see when a comm_msg
-    # comes across which indicates stream / display_data / error should be directed to the Output
-    # widget instead of the normal cell output. It will look up the comm_id in the CommManager and
-    # await these methods.
-    async def handle_stream(self, msg: messages.Stream):
-        self.state["outputs"].append(msg.content)
